/* $Id$ */
package saccubus.worker;

import static org.apache.commons.io.FilenameUtils.getBaseName;
import static org.apache.commons.lang.StringUtils.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingWorker;
import saccubus.conv.ConvertToVideoHook;
import saccubus.worker.profile.ConvertProfile;
import saccubus.worker.profile.ConvertProfile.HideCondition;
import saccubus.worker.profile.FfmpegProfile;
import saccubus.worker.profile.GeneralProfile;
import saccubus.worker.profile.OutputProfile;
import yukihane.inqubus.util.OutputNamePattern;
import yukihane.mediainfowrapper.Info;
import yukihane.mediainfowrapper.MediaInfo;
import yukihane.mediainfowrapper.Size;
import yukihane.swf.Cws2Fws;

/**
 * 動画を(コメント付きに)変換するワーカクラス.
 * @author yuki
 */
public class Convert extends SwingWorker<ConvertResult, ConvertProgress> {

    private static final Logger logger = Logger.getLogger(Convert.class.getName());
    private final ConvertProfile profile;
    private final File videoFile;
    private final File commentFile;

    /**
     * 変換ワーカコンストラクタ.
     * @param profile 変換用プロファイル.
     * @param video 変換元動画.
     * @param comment 変換元コメント. コメントを付与しない場合はnull.
     * @param output 変換後出力動画.
     * @throws IOException 変換失敗.
     */
    public Convert(ConvertProfile profile, File video, File comment) {
        this.profile = profile;
        this.videoFile = video;
        this.commentFile = comment;
    }

    @Override
    protected ConvertResult doInBackground() throws Exception {
        if (!profile.isConvert()) {
            return new ConvertResult(true, "");
        }

        final GeneralProfile gene = profile.getGeneralProfile();
        final OutputProfile outprof = profile.getOutputProfile();
        final OutputNamePattern pattern = new OutputNamePattern(outprof.getFileName());
        final String id = outprof.getVideoId();
        pattern.setId(isNotEmpty(id) ? id : "");
        final String title = outprof.getTitile();
        pattern.setTitle(isNotEmpty(title) ? title : "");
        final String fileName = getBaseName(videoFile.getPath());
        pattern.setFileName(fileName);
        pattern.setReplaceFrom(gene.getReplaceFrom());
        pattern.setReplaceFrom(gene.getReplaceTo());
        final File outputFile = new File(outprof.getDir(), pattern.createFileName());

        File transformedComment = null;
        try {

            if (profile.isCommentOverlay()) {
                transformedComment = File.createTempFile("vhk", ".tmp", profile.getTempDir());
                final HideCondition hide = profile.getNgSetting();
                publish(new ConvertProgress("コメントの中間ファイルへの変換中"));
                ConvertToVideoHook.convert(commentFile, transformedComment, hide.getId(), hide.getWord());
            }

            checkStop();
            publish(new ConvertProgress("動画の変換を開始"));

            final int code = convert(transformedComment, outputFile);
            if (code != 0) {
                throw new IOException("ffmpeg実行失敗: " + outputFile.getPath());
            }
            publish(new ConvertProgress("変換が正常に終了しました。"));
            return new ConvertResult(true, outputFile.getName());
        } finally {
            if (transformedComment != null && transformedComment.exists()) {
                transformedComment.delete();
            }
        }
    }

    private int convert(File transformedComment, File outputFile) throws InterruptedException, IOException {
        File fwsFile = null;
        try {
            final File tmpCws = File.createTempFile("cws", ".swf", profile.getTempDir());
            fwsFile = Cws2Fws.createFws(videoFile, tmpCws);
            tmpCws.delete();
            final File target = (fwsFile != null) ? fwsFile : videoFile;

            final List<String> arguments = createArguments(target, transformedComment, outputFile);
            return executeFfmpeg(arguments);
        } finally {
            if (fwsFile != null && fwsFile.exists()) {
                fwsFile.delete();
            }
        }
    }

    private List<String> createArguments(final File targetVideoFile, File transformedComment, File output)
            throws IOException, UnsupportedEncodingException {
        final ConvertProfile prof = profile;
        final FfmpegProfile ffop = prof.getFfmpegOption();

        final List<String> cmdList = new ArrayList<String>();
        cmdList.add(prof.getFfmpeg().getPath());
        cmdList.add("-y");
        final String[] mainOptions = ffop.getMainOption().split(" +");
        for (String opt : mainOptions) {
            if (isNotBlank(opt)) {
                cmdList.add(opt);
            }
        }
        final String[] inOptions = ffop.getInOption().split(" +");
        for (String opt : inOptions) {
            if (isNotBlank(opt)) {
                cmdList.add(opt);
            }
        }
        cmdList.add("-i");
        cmdList.add(targetVideoFile.getPath());
        final String[] outOptions = ffop.getOutOption().split(" +");
        for (String opt : outOptions) {
            if (isNotBlank(opt)) {
                cmdList.add(opt);
            }
        }
        final Info info = MediaInfo.getInfo(new File("bin", "MediaInfo"), targetVideoFile);
        // 4:3 なら1.33, 16:9 なら1.76
        final boolean isHD = ((double) info.getWidth() / (double) info.getHeight() > 1.5);
        if (ffop.isResize()) {
            final Size scaled = (ffop.isAdjustRatio()) ? MediaInfo.adjustSize(info, ffop.getResizeWidth(), ffop.
                    getResizeHeight()) : new Size(info.getWidth(), info.getHeight());
            cmdList.add("-s");
            cmdList.add(scaled.getWidth() + "x" + scaled.getHeight());
        }
        final List<String> avfilterArgs = createAvfilterOptions(ffop.getAvfilterOption());
        if (!prof.isVhookDisabled()) {
            // TODO 引数冗長
            final String vhookArg = getVhookArg(prof, prof.isCommentOverlay(), transformedComment.getPath(), isHD);
            if (isNotBlank(vhookArg)) {
                avfilterArgs.add(vhookArg);
            }
        }
        if (!avfilterArgs.isEmpty()) {
            cmdList.add("-vfilters");
            final String args = "\"" + join(avfilterArgs, ", ") + "\"";
            cmdList.add(args);
        }
        cmdList.add(output.getPath());
        final StringBuilder argMsg = new StringBuilder();
        argMsg.append("arg:");
        for (String s : cmdList) {
            argMsg.append(" ").append(s);
        }
        logger.log(Level.INFO, argMsg.toString());
        return cmdList;
    }

    private int executeFfmpeg(final List<String> cmdList) throws InterruptedException, IOException {
        Process process = null;
        try {
            logger.log(Level.INFO, "Processing FFmpeg...");
            process = Runtime.getRuntime().exec(cmdList.toArray(new String[0]));
            BufferedReader ebr = new BufferedReader(new InputStreamReader(
                    process.getErrorStream()));
            String msg;
            while ((msg = ebr.readLine()) != null) {
                if (msg.startsWith("frame=")) {
                    publish(new ConvertProgress(msg));
                } else if (!msg.endsWith("No accelerated colorspace conversion found")) {
                    logger.log(Level.INFO, msg);
                }

                checkStop();
            }

            process.waitFor();
            return process.exitValue();
        } finally {
            // TODO 正常終了した場合もdestroyしていいのか？
            if (process != null) {
                process.destroy();
            }
        }
    }

    private static List<String> createAvfilterOptions(String avfilterOption) {
        final List<String> avfilterArgs = new ArrayList<String>();
        if (isNotBlank(avfilterOption)) {
            avfilterArgs.add(avfilterOption);
        }
        return avfilterArgs;
    }

    private static String getVhookArg(ConvertProfile prof, boolean addComment, String commPath, boolean isHD) throws
            UnsupportedEncodingException {
        StringBuilder sb = new StringBuilder();
        sb.append("vhext=");
        sb.append(prof.getVhook().getPath().replace("\\", "/"));
        if (addComment) {
            sb.append("|");
            sb.append("--data-user:");
            sb.append(URLEncoder.encode(commPath.replace("\\", "/"), "Shift_JIS"));
        }
        sb.append("|");
        sb.append("--font:");
        sb.append(URLEncoder.encode(
                prof.getFont().getPath().replace("\\", "/"), "Shift_JIS"));
        sb.append("|");
        sb.append("--font-index:");
        sb.append(prof.getFontIndex());
        sb.append("|");
        sb.append("--show-user:");
        sb.append(prof.getMaxNumOfComment());
        sb.append("|");
        sb.append("--shadow:");
        sb.append(prof.getShadowIndex());
        sb.append("|");
        if (prof.isShowConverting()) {
            sb.append("--enable-show-video");
            sb.append("|");
        }
        if (!prof.isDisableFontSizeArrange()) {
            sb.append("--enable-fix-font-size");
            sb.append("|");
        }
        if (prof.isCommentOpaque()) {
            sb.append("--enable-opaque-comment");
            sb.append("|");
        }
        if (isHD) {
            sb.append("--aspect-mode:1");
            sb.append("|");
        }
        return sb.toString();
    }

    protected void checkStop() throws InterruptedException {
        if (Thread.interrupted()) {
            throw new InterruptedException("中止要求を受け付けました");
        }
    }
}
